@@ -30,7 +30,11 @@ class Production(Workflow, ModelSQL, ModelView):
             'readonly': ~Eval('state').in_(['request', 'draft']),
             },
         depends=['state'])
-    effective_date = fields.Date('Effective Date', readonly=True)
+    effective_date = fields.Date('Effective Date',
+        states={
+            'readonly': Eval('state').in_(['cancel', 'done']),
+            },
+        depends=['state'])
     company = fields.Many2One('company.company', 'Company', required=True,
         states={
             'readonly': ~Eval('state').in_(['request', 'draft']),
@@ -44,7 +48,6 @@ class Production(Workflow, ModelSQL, ModelView):
             'readonly': (~Eval('state').in_(['request', 'draft'])
                 | Eval('inputs', True) | Eval('outputs', True)),
             },
-        on_change=['warehouse'],
         depends=['state'])
     location = fields.Many2One('stock.location', 'Location', required=True,
         domain=[
@@ -61,8 +64,7 @@ class Production(Workflow, ModelSQL, ModelView):
             ],
         states={
             'readonly': ~Eval('state').in_(['request', 'draft']),
-            },
-        on_change=BOM_CHANGES)
+            })
     bom = fields.Many2One('production.bom', 'BOM',
         domain=[
             ('output_products', '=', Eval('product', 0)),
@@ -71,11 +73,10 @@ class Production(Workflow, ModelSQL, ModelView):
             'readonly': ~Eval('state').in_(['request', 'draft']),
             'invisible': ~Eval('product'),
             },
-        on_change=BOM_CHANGES,
         depends=['product'])
     uom_category = fields.Function(fields.Many2One(
-            'product.uom.category', 'Uom Category',
-            on_change_with=['product']), 'on_change_with_uom_category')
+            'product.uom.category', 'Uom Category'),
+        'on_change_with_uom_category')
     uom = fields.Many2One('product.uom', 'Uom',
         domain=[
             ('category', '=', Eval('uom_category')),
@@ -85,10 +86,9 @@ class Production(Workflow, ModelSQL, ModelView):
             'required': Bool(Eval('bom')),
             'invisible': ~Eval('product'),
             },
-        on_change=BOM_CHANGES,
         depends=['uom_category'])
-    unit_digits = fields.Function(fields.Integer('Unit Digits',
-            on_change_with=['uom']), 'on_change_with_unit_digits')
+    unit_digits = fields.Function(fields.Integer('Unit Digits'),
+        'on_change_with_unit_digits')
     quantity = fields.Float('Quantity',
         digits=(16, Eval('unit_digits', 2)),
         states={
@@ -96,10 +96,9 @@ class Production(Workflow, ModelSQL, ModelView):
             'required': Bool(Eval('bom')),
             'invisible': ~Eval('product'),
             },
-        on_change=BOM_CHANGES,
         depends=['unit_digits'])
     cost = fields.Function(fields.Numeric('Cost', digits=(16, 4),
-            readonly=True, on_change_with=['inputs']), 'get_cost')
+            readonly=True), 'get_cost')
     inputs = fields.One2Many('stock.move', 'production_input', 'Inputs',
         domain=[
             ('from_location', 'child_of', [Eval('warehouse')], 'parent'),
@@ -134,8 +133,9 @@ class Production(Workflow, ModelSQL, ModelView):
     def __setup__(cls):
         super(Production, cls).__setup__()
         cls._error_messages.update({
-                'missing_cost': ('Production "%s" misses costs on '
-                    'some of its outputs.'),
+                'uneven_costs': ('The costs of the outputs (%(outputs)s) of '
+                    'production "%(production)s" do not match the cost of the '
+                    'production (%(costs)s).')
                 })
         cls._transitions |= set((
                 ('request', 'draft'),
@@ -285,7 +285,7 @@ class Production(Workflow, ModelSQL, ModelView):
             values = self._explode_move_values(storage_location, self.location,
                 self.company, input_, quantity)
             if values:
-                inputs['add'].append(values)
+                inputs['add'].append((-1, values))
                 quantity = Uom.compute_qty(input_.uom, quantity,
                     input_.product.default_uom)
                 changes['cost'] += (Decimal(str(quantity)) *
@@ -305,9 +305,10 @@ class Production(Workflow, ModelSQL, ModelView):
                     values['unit_price'] = Decimal(
                         changes['cost'] / Decimal(str(quantity))
                         ).quantize(Decimal(str(10 ** -digits[1])))
-                outputs['add'].append(values)
+                outputs['add'].append((-1, values))
         return changes
 
+    @fields.depends('warehouse')
     def on_change_warehouse(self):
         changes = {
             'location': None,
@@ -316,6 +317,7 @@ class Production(Workflow, ModelSQL, ModelView):
             changes['location'] = self.warehouse.production_location.id
         return changes
 
+    @fields.depends(*BOM_CHANGES)
     def on_change_product(self):
         result = {}
         if self.product:
@@ -337,21 +339,26 @@ class Production(Workflow, ModelSQL, ModelView):
         result.update(self.explode_bom())
         return result
 
+    @fields.depends('product')
     def on_change_with_uom_category(self, name=None):
         if self.product:
             return self.product.default_uom.category.id
 
+    @fields.depends('uom')
     def on_change_with_unit_digits(self, name=None):
         if self.uom:
             return self.uom.digits
         return 2
 
+    @fields.depends(*BOM_CHANGES)
     def on_change_bom(self):
         return self.explode_bom()
 
+    @fields.depends(*BOM_CHANGES)
     def on_change_uom(self):
         return self.explode_bom()
 
+    @fields.depends(*BOM_CHANGES)
     def on_change_quantity(self):
         return self.explode_bom()
 
@@ -363,8 +370,11 @@ class Production(Workflow, ModelSQL, ModelView):
             else:
                 cost_price = input_.product.cost_price
             cost += (Decimal(str(input_.internal_quantity)) * cost_price)
-        return cost
 
+        digits = self.__class__.cost.digits
+        return cost.quantize(Decimal(str(10 ** -digits[1])))
+
+    @fields.depends('inputs')
     def on_change_with_cost(self):
         Uom = Pool().get('product.uom')
 
@@ -448,7 +458,11 @@ class Production(Workflow, ModelSQL, ModelView):
             cost_price += (Decimal(str(output.quantity))
                 * output.unit_price)
         if not self.company.currency.is_zero(self.cost - cost_price):
-            self.raise_user_error('missing_cost', (self.rec_name,))
+            self.raise_user_error('uneven_costs', {
+                    'production': self.rec_name,
+                    'costs': self.cost,
+                    'outputs': cost_price,
+                    })
 
     @classmethod
     def create(cls, vlist):
@@ -465,9 +479,9 @@ class Production(Workflow, ModelSQL, ModelView):
         return productions
 
     @classmethod
-    def write(cls, productions, values):
-        super(Production, cls).write(productions, values)
-        for production in productions:
+    def write(cls, *args):
+        super(Production, cls).write(*args)
+        for production in sum(args[::2], []):
             production._set_move_planned_date()
 
     def _get_move_planned_date(self):
@@ -537,7 +551,7 @@ class Production(Workflow, ModelSQL, ModelView):
         Move = pool.get('stock.move')
         Date = pool.get('ir.date')
         Move.do([m for p in productions for m in p.outputs])
-        cls.write(productions, {
+        cls.write([p for p in productions if not p.effective_date], {
                 'effective_date': Date.today(),
                 })
 
